@@ -55,6 +55,11 @@
                  (in? y (list "framed-title" "ridged-paper" "framed-session"))))
   #t)
 
+(tm-define (screens-buffer?)
+  (with t (buffer-tree)
+    (and (tree-is? t 'document)
+         (tree-is? t :last 'screens))))
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Dynamic movements for fold tags and switches
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -126,7 +131,9 @@
   (:require (toggle-context? t))
   (with i (if (toggle-first-context? t) 1 0)
     (variant-set t (ahash-ref alternate-table (tree-label t)))
-    (tree-go-to t i :start)))
+    (tree-go-to t i :start)
+    (when (> i 0)
+      (players-set-elapsed (tree-ref t i) 0.0))))
 
 (tm-define (dynamic-extremal t forwards?)
   (:require (toggle-context? t))
@@ -212,15 +219,18 @@
   (set! i (switch-index t i))
   (if (null? args) (set! args '(:start)))
   (when (switch-valid-child? t i)
-    (switch-select t i)
-    (apply tree-go-to (cons* t i 0 args))))
+    (with old (switch-index t)
+      (switch-select t i)
+      (apply tree-go-to (cons* t i 0 args))
+      (when (> i old)
+	(players-set-elapsed (tree-ref t i) 0.0)))))
 
-(tm-define (switch-insert-at t i)
+(tm-define (switch-insert-at t i what)
   (set! i (if (== i :end) (tree-arity t) (switch-index t i)))
   (when (and (>= i 0) (<= i (tree-arity t)))
     (let* ((empty (if (tree-in? t (big-switch-tag-list)) '(document "") ""))
            (v (switch-index t :visible)))
-      (tree-insert! t i `((shown ,empty)))
+      (tree-insert! t i `((shown ,(or what empty))))
       (if (tree-in? t (alternative-tag-list))
           (switch-select t i)
           (switch-select t (+ v 1)))
@@ -242,7 +252,7 @@
 ;; Specific types of switches
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(define (alternative-context? t)
+(tm-define (alternative-context? t)
   (alternative-tag? (tree-label t)))
 
 (tm-define (switch-select t i)
@@ -291,7 +301,7 @@
 
 (tm-define (structured-insert-horizontal t forwards?)
   (:require (switch-context? t))
-  (switch-insert-at t (if forwards? :var-next :current)))
+  (switch-insert-at t (if forwards? :var-next :current) #f))
 
 (tm-define (structured-insert-vertical t downwards?)
   (:require (switch-context? t))
@@ -367,11 +377,27 @@
        (tree-atomic? (tree-ref t 1))
        (tree->number (tree-ref t 1))))
 
+(define (overlays-animate t old new)
+  (when (tree-compound? t)
+    (when (and (overlay-visible? t new)
+               (not (overlay-visible? t old))
+	       (> new old))
+      (players-set-elapsed t 0.0))
+    (for-each (cut overlays-animate <> old new)
+              (tree-children t))))
+
+(define (overlays-set t new)
+  (with old (tree->number (tree-ref t 0))
+    (when (!= new old)
+      (tree-set t 0 (number->string new))
+      (graphics-update-proviso (tree-ref t 2) (number->string new))
+      (overlays-animate t old new))))
+
 (tm-define (overlays-switch-to t i)
   (when (overlays-context? t)
     (let* ((tot (tree->number (tree-ref t 1)))
            (nxt (min (max i 1) tot)))
-      (tree-set t 0 (number->string nxt)))))
+      (overlays-set t nxt))))
 
 (tm-define (overlay-current t)
   (and-with p (tree-search-upwards t overlays-context?)
@@ -381,25 +407,64 @@
   (and-with p (tree-search-upwards t overlays-context?)
     (overlays-arity p)))
 
+(define (overlay-satisfies-proviso? t i pos)
+  (cond ((>= pos (- (tree-arity t) 1)) #f)
+        ((tm-equal? (tree-ref t pos) "proviso")
+         (and-with c (tree-ref t (+ pos 1))
+           (and (tree-in? c (nullary-overlay-tag-list))
+                (== (tree-arity c) 1)
+                (and-with ref (tree->number (tree-ref c 0))
+                  (and (integer? ref)
+                       (cond ((tree-is? c 'show-always) #t)
+                             ((tree-is? c 'show-from)   (>= i ref))
+                             ((tree-is? c 'show-until)  (<= i ref))
+                             ((tree-is? c 'show-this)   (== i ref))
+                             ((tree-is? c 'show-other)  (!= i ref))
+                             (else #f)))))))
+        (else (overlay-satisfies-proviso? t i (+ pos 2)))))
+
+(define (overlay-visible-sub? t i)
+  (cond ((tree-atomic? t) #t)
+	((or (overlay-context? t)
+	     (tree-in? t '(with anim-static anim-morph)))
+	 (overlay-visible? t i))
+	(else (forall? (cut overlay-visible-sub? <> i)
+		       (tree-children t)))))
+
+(define (overlay-frame-visible? f i)
+  (and (tree-func? f 'tuple 2)
+       (overlay-visible-sub? (tree-ref f 1) i)))
+
 (tm-define (overlay-visible? t i)
-  (and (overlay-context? t)
-       (with ref (tree->number (tree-ref t 0))
-         (cond ((not (integer? ref)) #f)
-               ((tree-is? t 'overlay-from)    (>= i ref))
-               ((tree-is? t 'overlay-until)   (<= i ref))
-               ((tree-is? t 'overlay-this)    (== i ref))
-               ((tree-is? t 'overlay-other)   (!= i ref))
-               ((tree-is? t 'alternate-from)  (>= i ref))
-               ((tree-is? t 'alternate-until) (<= i ref))
-               ((tree-is? t 'alternate-this)  (== i ref))
-               ((tree-is? t 'alternate-other) (!= i ref))
-               (else #f)))))
+  (or (and (overlay-context? t)
+           (with ref (tree->number (tree-ref t 0))
+             (cond ((not (integer? ref)) #f)
+                   ((tree-is? t 'show-always)     #t)
+                   ((tree-is? t 'show-from)       (>= i ref))
+                   ((tree-is? t 'show-until)      (<= i ref))
+                   ((tree-is? t 'show-this)       (== i ref))
+                   ((tree-is? t 'show-other)      (!= i ref))
+                   ((tree-is? t 'overlay-from)    (>= i ref))
+                   ((tree-is? t 'overlay-until)   (<= i ref))
+                   ((tree-is? t 'overlay-this)    (== i ref))
+                   ((tree-is? t 'overlay-other)   (!= i ref))
+                   ((tree-is? t 'alternate-from)  (>= i ref))
+                   ((tree-is? t 'alternate-until) (<= i ref))
+                   ((tree-is? t 'alternate-this)  (== i ref))
+                   ((tree-is? t 'alternate-other) (!= i ref))
+                   (else #f))))
+      (and (tree-is? t 'with)
+           (overlay-satisfies-proviso? t i 0))
+      (and (tree-in? t '(anim-static anim-dynamic))
+	   (tree-is? t 0 'morph)
+	   (list-and (map (cut overlay-frame-visible? <> i)
+			  (tree-children (tree-ref t 0)))))))
 
 (tm-define (dynamic-extremal t forwards?)
   (:require (overlays-context? t))
   (let* ((tot (tree->number (tree-ref t 1)))
          (nxt (if forwards? tot 1)))
-    (tree-set t 0 (number->string nxt))
+    (overlays-set t nxt)
     (when (not (tree-innermost overlay-context?))
       (tree-go-to t 2 :start))))
 
@@ -409,7 +474,7 @@
          (tot (tree->number (tree-ref t 1)))
          (inc (if forwards? 1 -1))
          (nxt (min (max (+ cur inc) 1) tot)))
-    (tree-set t 0 (number->string nxt))
+    (overlays-set t nxt)
     (when (not (tree-innermost overlay-context?))
       (tree-go-to t 2 :start))))
 
@@ -455,7 +520,9 @@
          (tot (tree->number (tree-ref t 1)))
          (ins (if forwards? (+ cur 1) cur)))
     (overlay-renumber (tree-ref t 2) ins tot 1)
-    (if forwards? (tree-set t 0 (number->string ins)))
+    (when forwards?
+      (tree-set t 0 (number->string ins))
+      (graphics-update-proviso (tree-ref t 2) (number->string ins)))
     (tree-set t 1 (number->string (+ tot 1)))))
 
 (tm-define (structured-remove-horizontal t forwards?)
@@ -465,7 +532,9 @@
          (del (if forwards? cur (- cur 1))))
     (when (and (> del 0) (> tot 1))
       (overlay-renumber (tree-ref t 2) del tot -1)
-      (if (not forwards?) (tree-set t 0 (number->string del)))
+      (when (not forwards?)
+        (tree-set t 0 (number->string del))
+        (graphics-update-proviso (tree-ref t 2) (number->string del)))
       (tree-set t 1 (number->string (- tot 1))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -596,7 +665,7 @@
                ((== mode :last)
                 (dynamic-extremal t #t))))))
 
-(define (dynamic-operate t mode)
+(tm-define (dynamic-operate t mode)
   (when (tree-compound? t)
     (cond ((tree-is? t 'traversed)
            (when (!= mode :var-last)
@@ -623,19 +692,19 @@
 (define (dynamic-first-alternative-list l)
   (if (null? l) #f
       (with r (dynamic-first-alternative (car l))
-	(or r (dynamic-first-alternative-list (cdr l))))))
+        (or r (dynamic-first-alternative-list (cdr l))))))
 
 (define (dynamic-first-alternative t)
   (cond ((and (alternative-context? t) (> (tm-arity t) 1)) t)
-	((overlays-context? t)
+        ((overlays-context? t)
          (tree-set! t 0 "1")
          (tree-assign-node! t 'overlays-range)
          t)
         ((and (tree-is? t 'overlays-range)
               (< (tree->number (tree-ref t 0))
                  (tree->number (tree-ref t 1)))) t)
-	((not (tm-compound? t)) #f)
-	(else (dynamic-first-alternative-list (tm-children t)))))
+        ((not (tm-compound? t)) #f)
+        (else (dynamic-first-alternative-list (tm-children t)))))
 
 (define (dynamic-alternative-keep-first slide)
   (and-with t (dynamic-first-alternative slide)
@@ -658,29 +727,74 @@
     (tree-assign-node! t 'slide))
   (if (and (tm-func? t 'slide 1) (dynamic-first-alternative t))
       (let* ((p (tree-up t))
-	     (i (tree-index t)))
-	(tree-insert p (+ i 1) (list (tree-copy t)))
-	(let* ((v  (tree-ref p i))
-	       (w  (tree-ref p (+ i 1))))
-	  (dynamic-alternative-keep-first v)
-	  (dynamic-alternative-keep-other w)
-	  (dynamic-make-slide v)
-	  (dynamic-make-slide w)))
+             (i (tree-index t)))
+        (tree-insert p (+ i 1) (list (tree-copy t)))
+        (let* ((v  (tree-ref p i))
+               (w  (tree-ref p (+ i 1))))
+          (dynamic-alternative-keep-first v)
+          (dynamic-alternative-keep-other w)
+          (dynamic-make-slide v)
+          (dynamic-make-slide w)))
       (dynamic-operate (tree-ref t 0) :var-expand)))
 
+(define (keep-shown! t)
+  (if (alternative-context? t)
+      (map ;; Prune hidden children
+       (lambda (i) (if (tm-func? (tree-ref t i) 'hidden)
+                       (tree-remove! t i 1)))
+       (reverse (.. 0 (tm-arity t)))))
+  (if (and (tm-compound? t) (> (tm-arity t) 0))
+      (for-each keep-shown! (tree-children t))))
+
+(define (create-slide scr)
+  (if (not (tree? scr)) (tree 'slide '(document "")) ;; just in case
+      (with t (tree-copy scr)
+        (keep-shown! t)
+        (tree-assign-node! t 'slide)
+        t)))
+
+(define (process-screen scr)
+  (cons (create-slide scr)
+        (begin
+          (dynamic-traverse-buffer :next)
+          (if (tm-func? scr 'hidden) '()
+              (process-screen scr)))))
+
+(define (list->tree label l)
+  (tree-insert (tree label) 0 l))
+
+(define (screens->slides t)
+  (if (not (tm-func? t 'screens)) (tree 'document "")
+      (with f (lambda (scr) (list->tree 'document (process-screen scr)))
+        ;; (system-wait "Generating slides" "please wait") ;crashes if printing
+        ;; Insert fake screen at the end
+        (tree-insert! t (tree-arity t) 
+                      (list (tree 'hidden '(document ""))))
+        (dynamic-operate-on-buffer :first)
+        ;; Notice that we don't process the last (fake) screen
+        (list->tree 'screens (map f (cDr (tree-children t)))))))
+
 (tm-define (dynamic-make-slides)
-  (add-style-package "slides")
   (init-default "page-medium" "page-type" "page-width" "page-height"
                 "page-width-margin" "page-height-margin"
-                "page-odd" "page-even" "page-right"
-                "par-width" "page-odd-shift" "page-even-shift"
-                "page-top" "page-bot" "page-height-margin")
-  (let* ((t (buffer-tree))
-         (c (tree-children t)))
-    (when (and (tm-func? t 'document) (switch-context? (cAr c)))
-      (tree-assign-node (cAr c) 'document)
-      (tree-set! t `(document ,@(cDr c) ,@(tree-children (cAr c))))
-      (for-each dynamic-make-slide (tree-children t)))))
+                ;;"page-odd" "page-even" "page-right"
+                ;;"par-width" "page-odd-shift" "page-even-shift"
+                ;;"page-top" "page-bot"
+                )
+  (add-style-package "slides")
+  (if (preference-on? "texmacs->pdf:expand slides")
+      (let* ((t (buffer-tree))
+             (c (tree-children t)))
+        (when (and (tm-func? t 'document) (switch-context? (cAr c)))
+          (tree-assign-node (cAr c) 'document)
+          (tree-set! t `(document ,@(cDr c) ,@(tree-children (cAr c))))
+          ;; (system-wait "Generating slides" "please wait") ;crashes if printing
+          (for-each dynamic-make-slide (tree-children t))))
+      (with l (select (buffer-tree) '(screens))
+        (and (nnull? l)
+             (let* ((scrns (car l))
+                    (slides (screens->slides scrns)))
+               (tree-set! scrns slides))))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Global filtering of switches
@@ -778,7 +892,7 @@
          (nxt (min (max (+ cur inc) 1) tot)))
     (and (!= nxt cur)
          (begin
-           (tree-set t 0 (number->string nxt))
+           (overlays-set t nxt)
            (when (not (tree-innermost overlay-context?))
              (tree-go-to t 2 :start))
            #t))))
@@ -798,7 +912,7 @@
        (or (dynamic-traverse (car l) mode)
            (dynamic-traverse-list (cdr l) mode))))
 
-(define (dynamic-traverse t mode)
+(tm-define (dynamic-traverse t mode)
   (cond ((tree-atomic? t) #f)
         ((tree-is? t 'traversed)
          (dynamic-traverse-traversed t mode))
@@ -832,24 +946,6 @@
 (tm-define (dynamic-traverse-buffer mode)
   (dynamic-traverse (buffer-tree) mode))
 
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;; Specific navigation for 'screens' switch
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-
-(tm-define (screens-switch-to which)
-  (:secure #t)
-  (and-with t (tree-innermost 'screens)
-    (switch-to t which :start)))
-
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-;; Entering slide titles
-;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
-
-(tm-define (kbd-enter t shift?)
-  (:require (tree-is? t 'tit))
-  (tree-go-to t :end)
-  (insert-return))
-
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Specific behaviour for switches inside list environments
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -869,5 +965,296 @@
   (:require (switch-list-context? t))
   (if shift? (make-return-after)
       (begin
-        (switch-insert-at t :var-next)
+        (switch-insert-at t :var-next #f)
         (make-item))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Entering slide titles
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(tm-define (kbd-enter t shift?)
+  (:require (tree-is? t 'tit))
+  (tree-go-to t :end)
+  (insert-return))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Specific routines for 'screens' switch
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(tm-define (slideshow-context? t)
+  (and (tree-is? t 'slideshow)
+       (== (tree-arity t) 1)
+       (tree-is? (tree-ref t 0) 'document)))
+
+(tm-define (screens-context? t)
+  (or (tree-is? t 'screens)
+      (slideshow-context? t)))
+
+(tm-define (nr-pages)
+  (:require (tree-innermost screens-context?))
+  (with t* (tree-innermost screens-context?)
+    (with t (if (slideshow-context? t*) (tree-ref t* 0) t*)
+      (tree-arity t))))
+
+(tm-define (screens-switch-to which)
+  (:secure #t)
+  (and-with t (tree-innermost 'screens)
+    (switch-to t which :start)))
+
+(tm-define (screens-show-all)
+  (and-with t (tree-innermost 'screens)
+    (for (c (tree-children t))
+      (when (tree-in? c '(hidden shown))
+        (tree-assign-node! c 'slide)))
+    (tree-assign-node! t 'document)
+    (tree-insert-node! t 0 '(slideshow))))
+
+(tm-define (screens-show-this)
+  (and-with t (tree-innermost 'slideshow)
+    (when (and (== (tree-arity t) 1) (tree-is? (tree-ref t 0) 'document))
+      (tree-remove-node! t 0)
+      (tree-assign-node! t 'screens)
+      (for (c (tree-children t))
+	(when (tree-func? c 'slide)
+	  (tree-assign-node! c (if (cursor-inside? c) 'shown 'hidden)))))))
+
+(define (expand-slides? s)
+  (in? s (list "paper" "book" "panorama")))
+
+(tm-define (init-page-rendering s)
+  (:require (or (inside? 'screens) (inside? 'slideshow)))
+  (with o (get-init-page-rendering)
+    (when (and (expand-slides? s) (not (expand-slides? o)))
+      (screens-show-all))
+    (when (and (not (expand-slides? s)) (expand-slides? o))
+      (screens-show-this))
+    (former s)))
+
+(tm-define (slide-get-switch t)
+  (if (slideshow-context? t) (tree-ref t 0) t))
+
+(tm-define (slide-get-document t)
+  (cond ((not (tree? t)) #f)
+        ((slideshow-context? t)
+         (slide-get-document (tree-ref t 0 :down :down)))
+        ((screens-context? t)
+         (slide-get-document (tree-ref t :down :down)))
+        ((tree-in? t '(shown hidden screen slide))
+         (slide-get-document (tree-ref t 0)))
+        ((tree-in? t '(with with-screen-color))
+         (slide-get-document (tree-ref t :last)))
+        ((tree-is? t 'document) t)
+        (else #f)))
+
+(define (slide-get-bg-color-bis t)
+  (cond ((tree-is? t :up 'with)
+         (slide-set-bg-color-bis (tree-up t)))
+        ((tree-is? t :up 'with-screen-color)
+         (tree-ref t :up 0))
+        (else "")))
+
+(tm-define (slide-get-bg-color)
+  (and-with t (tree-innermost screens-context?)
+    (with u (slide-get-document t)
+      (slide-get-bg-color-bis u))))
+
+(define (slide-set-bg-color-bis t col)
+  (cond ((tree-is? t :up 'with)
+         (slide-set-bg-color-bis (tree-up t) col))
+        ((tree-is? t :up 'with-screen-color)
+         (if (== col "")
+             (tree-set t :up t)
+             (tree-set t :up 0 col)))
+        ((!= col "")
+         (tree-set t `(with-screen-color ,col ,t)))))
+
+(tm-define (slide-set-bg-color col)
+  (and-with t (tree-innermost screens-context?)
+    (with u (slide-get-document t)
+      (slide-set-bg-color-bis u col))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Editing slideshows in expanded form
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(tm-define (screens-switch-to which)
+  (:require (inside? 'slideshow))
+  (and-with t (tree-innermost 'slideshow)
+    (when (and (== (tree-arity t) 1) (tree-is? (tree-ref t 0) 'document))
+      (set! t (tree-ref t 0))
+      (when (== which :first) (set! which 0))
+      (when (== which :last) (set! which (- (tree-arity t) 1)))
+      (tree-go-to t which :start))))
+
+(tm-define (structured-horizontal? t)
+  (:require (slideshow-context? t))
+  #t)
+
+(tm-define (focus-can-insert? t)
+  (:require (slideshow-context? t))
+  #t)
+
+(tm-define (focus-can-remove? t)
+  (:require (slideshow-context? t))
+  #t)
+
+(tm-define (structured-insert-horizontal t forwards?)
+  (:require (slideshow-context? t))
+  (set! t (tree-ref t 0))
+  (with i (+ (tree-index (tree-down t)) (if forwards? 1 0))
+    (tree-insert! t i '((slide (document ""))))
+    (tree-go-to t i 0 0 0)))
+
+(tm-define (structured-remove-horizontal t forwards?)
+  (:require (slideshow-context? t))
+  (set! t (tree-ref t 0))
+  (let* ((i (- (tree-index (tree-down t)) (if forwards? 0 1)))
+         (n (tree-arity t)))
+    (when (and (>= i 0) (> n 1))
+      (tree-remove! t i 1)
+      (tree-go-to t (min i (- n 1)) (if forwards? :start :end)))))
+
+(define (slide-range)
+  (and-with t (tree-innermost slideshow-context?)
+    (let* ((p (tree->path (tree-ref t 0)))
+           (p1 (selection-get-start))
+           (p2 (selection-get-end)))
+      (and (list-starts? p1 p)
+           (list-starts? p2 p)
+           (let* ((i1 (list-ref p1 (length p)))
+                  (i2 (list-ref p2 (length p)))
+                  (s1 (tm-ref t 0 i1))
+                  (s2 (tm-ref t 0 i2)))
+             (and s1 s2 (< i1 i2)
+                  (tree-is? s1 'slide) (tree-is? s2 'slide)
+                  (< (+ (- i2 i1) 1) (tree-arity (tree-ref t 0)))
+                  (list i1 i2)))))))
+
+(tm-define (clipboard-cut which)
+  (:require (and (inside? 'slideshow)
+                 (slide-range)))
+  (with (i1 i2) (slide-range)
+    (with t (tree-innermost slideshow-context?)
+      (clipboard-set which (selection-tree))
+      (tree-remove (tree-ref t 0) i1 (+ (- i2 i1) 1))
+      (if (>= i1 (tree-arity (tree-ref t 0)))
+          (tree-go-to t 0 (- i1 1) :end)
+          (tree-go-to t 0 i1 :start)))))
+
+(define (slides? t)
+  (and (tree-is? t 'tuple)
+       (> (tree-arity t) 2)
+       (== (tree-ref t 0) (string->tree "texmacs"))
+       (with u (tree-ref t 1)
+         (and (tree-is? u 'document)
+              (> (tree-arity u) 1)
+              (list-and (map (cut tree-is? <> 'slide) (tree-children u)))))))
+
+(tm-define (clipboard-paste which)
+  (:require (and (inside? 'slideshow)
+                 (slides? (clipboard-get which))))
+  (let* ((t (tree-innermost slideshow-context?))
+         (i (tree-index (tree-down (tree-ref t 0))))
+         (ins (tree-ref (clipboard-get which) 1)))
+    (tree-insert (tree-ref t 0) i (tree-children ins))))
+
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+;; Customizations for graphical screens
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(tm-define (in-screens-graphics?)
+  (and (inside? 'gr-screen) (in-graphics?)))
+
+(define (innermost-gr-screen)
+  (and-with t (tree-innermost 'screens)
+    (and-with s (slide-get-document t)
+      (or (and (tree-is? s 'document)
+               (>= (tree-arity s) 1)
+               (tree-func? (tree-ref s 0) 'gr-screen 1)
+               (tree-ref s 0))
+          (and (tree-is? s 'document)
+               (>= (tree-arity s) 2)
+               (tree-is? (tree-ref s 0) 'tit)
+               (tree-func? (tree-ref s 1) 'gr-screen 1)
+               (tree-ref s 1))))))
+
+(define (innermost-graphics-screen)
+  (and-with t (innermost-gr-screen)
+    (set! t (tree-ref t 0))
+    (while (or (tree-func? t 'document 1)
+               (tree-func? t 'gr-overlays 3)
+               (tree-is? t 'with))
+      (set! t (tm-ref t :last)))
+    (and (tree-is? t 'graphics) t)))
+
+(define (go-to-graphics)
+  (and-with t (innermost-graphics-screen)
+    (tree-go-to t 0 :start)
+    (graphics-reset-state)
+    (graphics-decorations-update)))
+
+(tm-define (dynamic-operate-on-buffer mode)
+  ;;(:require (in-screens-graphics?))
+  (former mode)
+  (go-to-graphics))
+
+(tm-define (dynamic-traverse-buffer mode)
+  ;;(:require (in-screens-graphics?))
+  (former mode)
+  (go-to-graphics))
+
+(tm-define (screens-switch-to which)
+  ;;(:require (in-screens-graphics?))
+  (former which)
+  (go-to-graphics))
+
+(define (extract-slide-template t)
+  (when (or (tree-func? t 'shown 1) (tree-func? t 'hidden 1))
+    (set! t (tree-ref t 0)))
+  (cond ((tree-func? t 'document 1)
+         `(document ,(extract-slide-template (tree-ref t 0))))
+        ((and (tree-func? t 'document)
+              (>= (tree-arity t) 2)
+              (tree-func? (tree-ref t 0) 'tit))
+         `(document ,(extract-slide-template (tree-ref t 0))
+                    ,(extract-slide-template (tree-ref t 1))))
+        ((and (tree-func? t 'document)
+              (>= (tree-arity t) 1))
+         `(document ,(extract-slide-template (tree-ref t 0))))
+        ((tree-is? t 'tit)
+         `(tit ""))
+        ((tree-func? t 'gr-screen 1)
+         `(gr-screen ,(extract-slide-template (tree-ref t 0))))
+        ((tree-func? t 'gr-overlays 3)
+         `(gr-overlays "1" "1" ,(extract-slide-template (tree-ref t 2))))
+        ((tree-in? t '(with with-screen-color))
+         (with l (tree-children t)
+           `(,(tree-label t) ,@(cDr l) ,(extract-slide-template (cAr l)))))
+        ((tree-is? t 'graphics)
+         `(graphics ""))
+        (else "")))
+
+(tm-define (structured-insert-horizontal t forwards?)
+  (:require (and (screens-context? t) (not (slideshow-context? t))))
+  (with tmpl (extract-slide-template (tree-down t))
+    (when (not (tree-multi-paragraph? (tm->tree tmpl)))
+      (set! tmpl `(document ,tmpl)))
+    (switch-insert-at t (if forwards? :var-next :current) tmpl)
+    (and-with s (tree-down (tree-innermost 'screens))
+      (when (or (tree-func? s 'shown 1) (tree-func? s 'hidden 1))
+        (set! s (tree-ref s 0)))
+      (cond ((and (tree-is? s 'document)
+                  (>= (tree-arity s) 2)
+                  (tree-func? (tree-ref s 0) 'tit 1))
+             (tree-go-to s 0 0 :start))
+            ((and (tree-is? s 'document)
+                  (>= (tree-arity s) 1)
+                  (tree-func? (tree-ref s 0) 'gr-screen 1))
+             (go-to-graphics))))))
+
+(tm-define (make-gr-overlays forwards?)
+  (and-with t (tree-innermost 'graphics #t)
+    (while (tree-is? t :up 'with)
+      (set! t (tree-up t)))
+    (when (not (overlays-context? t))
+      (tree-set! t `(gr-overlays ,(if forwards? "2" "1") "2" ,t)))))
